#  ==========================================================================
#  flattenvertices.rb
#  --------------------------------------------------------------------------
#
#  Module	:	BKF::PlanarVertices (Module for making things planar)
#  Author	:	Niall Campbell, 2011-08(-09)
#  Version	:	1.1.0 2011-08(-09)
#
#  --------------------------------------------------------------------------
#  ==========================================================================
#  --------------------------------------------------------------------------
#
#  Description	:	Simple script for making all vertices in the current selection
#					have a z value of 0.
#
#  Definition Of Use
#  -----------------
#
#  (1)	Flattening Vertices
#
#		First select all edges you want to flatten, then run Plugins->
#		Flatten Vertices.
#
#  --------------------------------------------------------------------------
#  ==========================================================================
#  --------------------------------------------------------------------------
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  --------------------------------------------------------------------------
#  ==========================================================================

module PlanarVertices

BKF_FLATTEN_TO_Z = 0
BKF_FLATTEN_BY_NORMAL = 1

	class FlattenVertices

		def initialize(type)
			model = Sketchup.active_model
			model.start_operation("Flatten Vertices",true)
			case type
				when BKF_FLATTEN_TO_Z
					vertices = self.get_vertices
					if vertices.length > 0
						self.flatten_vertices(vertices)
						Sketchup::set_status_text("Operation Complete: "+vertices.length.to_s+" vertices flattened",SB_PROMPT)
						model.commit_operation
					else
						puts "no vertices found"
						Sketchup::set_status_text("Operation Aborted: No vertices found",SB_PROMPT)
						model.abort_operation
					end #if
				when BKF_FLATTEN_BY_NORMAL
					faces = self.get_faces
					#border_vertices = self.get_border_vertices(faces)
					#border_normals = self.get_border_normals(border_vertices,faces)
					average_normal = self.get_average_normal(faces)
					vertices = self.get_vertices
					if vertices.length > 0
						self.flatten_vertices(vertices,average_normal)
						Sketchup::set_status_text("Operation Complete: "+vertices.length.to_s+" vertices flattened",SB_PROMPT)
						model.commit_operation
					else
						puts "no vertices found"
						Sketchup::set_status_text("Operation Aborted: No vertices found",SB_PROMPT)
						model.abort_operation
					end #if
			end #case
		end #def
		
		def get_border_normals(border_vertices,faces)
			border_normals = Hash.new()
			faces_hash = Hash.new()
			
			faces.each{|face|
				faces_hash[face.entityID] = face
			}
		
			border_vertices.each{|vertexID,vertexOBJ|
				vertex_faces = vertexOBJ.faces
				infl_normals = Array.new()
				vertex_faces.each{|face|
					infl_faces << face if !faces_hash.has_key?(face.entityID)
				}
				av_normal = infl_faces[0].normal
				for i in 1...infl_faces.length
					av_normal = Geom::Vector3d.linear_combination(1, av_normal, 1, infl_faces[i].normal)
					av_normal.normalize!
				end #for
				border_normals[vertexID] = av_normal
			}
		end #def
		
		
		def get_border_vertices(faces)
			model = Sketchup.active_model
			sel_ents = model.selection
			edges_array = Array.new()
			faces_hash = Hash.new()
			vertices_hash = Hash.new()
			
			faces.each{|face|
				faces_hash[face.entityID] = face
			}
			
			sel_ents.each do |ent|
				edges_array << ent if ent.kind_of?(Sketchup::Edge)
			end
			edges_array.each {|ent|
				border = false
				#so to find the dominant border vertex we first need to
				#find the right faces?
				#if we use the selected border face's normal and its edge as
				#the axis of rotation, then rotate away from its free point to
				#the first normal we find, we should get the right face or group
				#of faces.
				
				#actually, all we need to check is if the face is more complex than
				#a triangle, if it is, the edge it has is the dominant normal
				
				ent.faces.each {|face|
					if !faces_hash.has_key?(face.entityID)
						
						border = true
						break
					end #if
				}
				if border
					vertices = ent.vertices
					vertices.each{|vertex|
					
						vertices_hash[vertex.entityID] = vertex
					}
				end #if
			}
			return vertices_hash
		end #def
		
		def get_average_normal(faces)
			av_normal = faces[0].normal
			normal_weight = faces[0].area
			for i in 1...faces.length
				av_normal = Geom::Vector3d.linear_combination(1, av_normal, faces[1].area.quo(normal_weight), faces[i].normal)
				av_normal.normalize!
				normal_weight += faces[i].area
			end #for
			return av_normal
		end #def
		
		def get_faces
			model = Sketchup.active_model
			sel_ents = model.selection
			faces_array = Array.new()
			for i in 0...sel_ents.length
				ent = sel_ents[i]
				faces_array << ent if ent.kind_of?(Sketchup::Face)
			end #for
			return faces_array
		end #def

		def get_vertices
			model = Sketchup.active_model
			sel_ents = model.selection
			vertices_hash = Hash.new()
			for i in 0...sel_ents.length
				ent = sel_ents[i]
				if ent.kind_of?(Sketchup::Edge)
					vertices = ent.vertices
					vertices.each {|vertex|
						vertices_hash[vertex.entityID] = vertex
					}
				end #if
			end #for
			return vertices_hash
		end #def

		def flatten_vertices(vertices,normal=false,border_vertices=false,border_normals=false)
		
			model = Sketchup.active_model
			ents = model.entities
			tick = 0
			vertex_length = vertices.length.to_s
			
			if normal
				arb_point = 0
				vertices.each {|vertexID,vertexOBJ| arb_point = vertexOBJ.position; break;}
				arb_plane = [arb_point, normal]
				###############
				#find highest, lowest, average vertex distance to plane
				
				distance = 0
				vertices.each{|vertexID, vertexOBJ|
					#work out the vertex's position relative to the direction
					#of the plane's normal
					point = vertexOBJ.position
					point_on_plane = point.project_to_plane arb_plane
					vector = point_on_plane.vector_to(point)
					if vector.valid?
						length = ((vector.samedirection? normal) ? 1 : -1) * vector.length
						distance += length
					end #if
				}
				av_distance = distance.quo(vertices.length)
				av_vector = normal.clone
				av_vector.length = av_distance
				av_point = arb_point.offset av_vector
				
				av_plane = [av_point, normal]
			else
				av_plane = [Geom::Point3d.new(0,0,0),Geom::Vector3d.new(0,0,1)]
			end #if
			################
			
			vertices.each {|vertexID, vertexOBJ|
				tick += 1
				msg_status="Moving Vertices: Vertex"+" "+tick.to_s+"/"+vertex_length
				Sketchup::set_status_text(msg_status,SB_PROMPT)
				
				point = vertexOBJ.position
				point_on_plane = point.project_to_plane av_plane
				vector = point.vector_to(point_on_plane)
				ents.transform_by_vectors([vertexOBJ],[vector]) if vector.valid?
			}

		end #def
	end #class

end #module

#############################################
# this part adds the plugin option to the plugin menu
if not file_loaded?(File.basename(__FILE__))
  
	### submenu added by bentleykf(niall.campbell@gmail.com)
	submenu = UI.menu("Plugins").add_submenu("Flatten Vertices")
	cmd = UI::Command.new("To Z=0") {
		PlanarVertices::FlattenVertices.new(0)
	}

	submenu.add_item(cmd)
	
	cmd = UI::Command.new("To Mean Normal") {
		PlanarVertices::FlattenVertices.new(1)
	}

	submenu.add_item(cmd)
end #if
file_loaded(File.basename(__FILE__))